'''
New-made module for \Python27\Lib\dis.py
'''

from dis import *

OPCODE_MAX = 147

def findlabels(code):
    """Detect all offsets in a byte code which are jump targets.
    Return the list of offsets.
    """
    labels = []
    extended_labels = {}
    n = len(code)
    i = inst_num = 0
    while i < n:
        try:
            c = code[i]
            op = ord(c)
            i = i+1
            if op >= HAVE_ARGUMENT and op <= OPCODE_MAX and i+1 < n: #fake opcode
                if not '<' in opname[op]: #fake opcode
                    try:
                        oparg = ord(code[i]) + ord(code[i+1])*256
                    except:
                        continue #fake arg
                    i = i+2
                    label = -1
                    if op in hasjrel:
                        label = i+oparg
                    elif op in hasjabs:
                        label = oparg
                    if label >= 0:
                        if label not in labels:
                            labels.append(label)
                        if label not in extended_labels:
                            extended_labels[label] = [inst_num,]
                        else:
                            extended_labels[label].append(inst_num)
        finally:
            inst_num = inst_num+1
    return labels, extended_labels

def findlinestarts(code):
    """Find the offsets in a byte code which are start of lines in the source.
    Generate pairs (offset, lineno) as described in Python/compile.c.
    """
    byte_increments = [ord(c) for c in code.co_lnotab[0::2]]
    line_increments = [ord(c) for c in code.co_lnotab[1::2]]

    lastlineno = None
    lineno = code.co_firstlineno
    addr = 0
    for byte_incr, line_incr in zip(byte_increments, line_increments):
        if byte_incr:
            if lineno != lastlineno:
                yield (addr, lineno)
                lastlineno = lineno
            addr += byte_incr
        lineno += line_incr
    if lineno != lastlineno:
        yield (addr, lineno)

def disassemble(co, lasti=-1):
    """Disassemble a code object."""
    code = co.co_code
    labels, _ = findlabels(code)
    linestarts = dict(findlinestarts(co))
    n = len(code)
    i = 0
    extended_arg = 0
    free = None
    while i < n:
        c = code[i]
        op = ord(c)
        if i in linestarts:
            if i > 0:
                print
            print "%3d" % linestarts[i],
        else:
            print '   ',
        if i == lasti: print '-->',
        else: print '   ',
        if i in labels: print '>>',
        else: print '  ',
        print repr(i).rjust(4),
        fake_op = False
        if op <= OPCODE_MAX: 
            ops = opname[op]
            if '<' in ops:
                fake_op = True
                ops = ops + ' FAKE!' #fake opcode
            print ops.ljust(20),
        else:
            fake_op = True
            print ('<'+repr(op)+'> '+'FAKE!').ljust(20), #fake opcode
        print ('['+code[i].encode('hex').upper()),
        i = i+1
        if op >= HAVE_ARGUMENT and not fake_op and i+1 < n:
            print (code[i].encode('hex').upper()+' '+code[i+1].encode('hex').upper()+']'),
            try:
                oparg = ord(code[i]) + ord(code[i+1])*256 + extended_arg
            except:
                continue #fake arg
            extended_arg = 0
            i = i+2
            if op == EXTENDED_ARG:
                extended_arg = oparg*65536L
            print repr(oparg).rjust(5),
            if op in hasconst:
                print '(' + (repr(co.co_consts[oparg]) if (oparg >=0 and oparg < len(co.co_consts)) else 'FAKE!') + ')',
            elif op in hasname:
                print '(' + (co.co_names[oparg] if (oparg >=0 and oparg < len(co.co_names)) else 'FAKE!') + ')',
            elif op in hasjrel:
                print '(to ' + repr(i + oparg) + ')',
            elif op in haslocal:
                print '(' + (co.co_varnames[oparg] if (oparg >=0 and oparg < len(co.co_varnames)) else 'FAKE!') + ')',
            elif op in hascompare:
                print '(' + (cmp_op[oparg] if (oparg >=0 and oparg < len(cmp_op)) else 'FAKE!') + ')',
            elif op in hasfree:
                if free is None:
                    free = co.co_cellvars + co.co_freevars
                print '(' + (free[oparg] if (oparg >=0 and oparg < len(free)) else 'FAKE!') + ')',
        else:
            print '-- --]',
        print

def disassemble_string(code, lasti=-1, varnames=None, names=None,
                       constants=None):
    labels, _ = findlabels(code)
    n = len(code)
    i = 0
    while i < n:
        c = code[i]
        op = ord(c)
        if i == lasti: print '-->',
        else: print '   ',
        if i in labels: print '>>',
        else: print '  ',
        print repr(i).rjust(4),
        fake_op = False #<<< new code
        if op <= OPCODE_MAX: 
            ops = opname[op]
            if '<' in ops:
                fake_op = True
                ops = ops + ' FAKE!' #fake opcode
            print ops.ljust(20),
        else:
            fake_op = True
            print ('<'+repr(op)+'> '+'FAKE!').ljust(20), #fake opcode
        print ('['+code[i].encode('hex').upper()),
        i = i+1
        if op >= HAVE_ARGUMENT and not fake_op and i+1 < n:
            print (code[i].encode('hex').upper()+' '+code[i+1].encode('hex').upper()+']'),
            try:            
                oparg = ord(code[i]) + ord(code[i+1])*256
            except:
                continue #fake arg
            i = i+2
            print repr(oparg).rjust(5),
            if op in hasconst:
                if constants:
                    print '(' + (repr(constants[oparg]) if (oparg >=0 and oparg < len(constants)) else 'FAKE!') + ')',
                else:
                    print '(%d)'%oparg,
            elif op in hasname:
                if names is not None:
                    print '(' + (names[oparg] if (oparg >=0 and oparg < len(names)) else 'FAKE!') + ')',
                else:
                    print '(%d)'%oparg,
            elif op in hasjrel:
                print '(to ' + repr(i + oparg) + ')',
            elif op in haslocal:
                if varnames:
                    print '(' + (varnames[oparg] if (oparg >=0 and oparg < len(varnames)) else 'FAKE!') + ')',
                else:
                    print '(%d)' % oparg,
            elif op in hascompare:
                print '(' + (cmp_op[oparg] if (oparg >=0 and oparg < len(cmp_op)) else 'FAKE!') + ')',
        else:
            print '-- --]',
        print